/*
  Copyright (c) 2013, 2014 Jouni Sirén
  Copyright (c) 2016 Genome Research Ltd.

  Author: Jouni Sirén <jouni.siren@iki.fi>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#ifndef _PDLTOPK_DOCLIST_H
#define _PDLTOPK_DOCLIST_H

#include <stack>

#include "pdltree.h"
#include "sadacount.h"


class PDLTopk
{
  public:
    typedef std::vector<Document>                      result_type;
    typedef std::vector<ScoredDoc>                     scored_type;
    typedef enum { st_unfinished, st_error, st_ok }    status_type;
    typedef enum { merge_brute, merge_lin, merge_exp } merge_type;
    typedef enum { ranked_or, ranked_and }             query_type;

    const static std::string EXTENSION; // .topk
    const static std::string MERGE_EXTENSION; // .topk_merge

    // If compress_sets is true, sets and grammar will be read from the files compressed by irepair.
    PDLTopk(const RLCSA& _rlcsa, const std::string& base_name, bool compress_sets, bool merge);
    ~PDLTopk();

    inline bool isOk() const { return (this->status == st_ok); }
    inline bool supportsMerge() const { return this->supports_merge; }

    void writeTo(const std::string& base_name) const;
    usint reportSize(bool include_freqs) const;

    // User must delete the result.
    result_type* query(const std::string& pattern, usint k, bool get_freqs) const;
    result_type* query(pair_type sa_range, usint k, bool get_freqs) const;
    result_type* query(const std::string& pattern, usint k, merge_type strategy) const;
    result_type* query(pair_type sa_range, usint k, merge_type strategy) const;

    // Simple multi-term tf-idf.
    std::vector<ScoredDoc>* query(const std::vector<std::string>& terms, usint k, query_type type, const SadaSparse& counter) const;

  private:
    const RLCSA&     rlcsa;
    status_type      status;
    bool             supports_merge;

    PDLTree*         tree;
    CSA::MultiArray* freqs;
    CSA::ReadBuffer* grammar;
    CSA::MultiArray* blocks;

    result_type* queryUnsafe(pair_type sa_range, usint k, bool get_freqs) const;
    result_type* queryUnsafe(pair_type sa_range, usint k, merge_type strategy) const;

    // k == 0 means all documents.
    void addBlock(result_type* results, usint block, usint k, bool get_freqs) const;

    // Try to add 'amount' documents from each block; return true if added at least one.
    bool addFromBlocks(scored_type& doc_buffer, usint amount,
      std::vector<CSA::MultiArray::Iterator*> block_iters,
      std::vector<CSA::MultiArray::Iterator*> freq_iters,
      std::vector<pair_type>& next_freqs,
      std::vector<std::stack<usint>>& buffers) const;

    // As above, but use tf-idf scoring.
    bool addFromBlocks(scored_type& doc_buffer, usint amount,
      std::vector<CSA::MultiArray::Iterator*> block_iters,
      std::vector<CSA::MultiArray::Iterator*> freq_iters,
      std::vector<pair_type>& next_freqs,
      std::vector<std::stack<usint>>& buffers,
      const std::vector<double>& idfs) const;

    // Get the next document. Use freq_iter == 0 to ignore frequencies.
    Document next(CSA::MultiArray::Iterator* block_iter, CSA::MultiArray::Iterator* freq_iter,
      pair_type& next_freqs, std::stack<usint>& buffer) const;

    // These are not allowed.
    PDLTopk();
    PDLTopk(const PDLTopk&);
    PDLTopk& operator = (const PDLTopk&);
};


#endif  // _PDLTOPK_DOCLIST_H
