/*
  Copyright (c) 2013, 2014 Jouni Sirén
  Copyright (c) 2016 Genome Research Ltd.

  Author: Jouni Sirén <jouni.siren@iki.fi>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include <iostream>
#include <stack>

#include "sadacount.h"

//--------------------------------------------------------------------------

const std::string SADACOUNT_EXTENSION = ".sadacount";
const std::string SADAEXPORT_EXTENSION = ".sadaexport";
const std::string SADAFILTER_EXTENSION = ".sadafilter";

//--------------------------------------------------------------------------

SadaSparse::SadaSparse(const RLCSA& _rlcsa, const std::string& base_name) :
  rlcsa(_rlcsa), status(st_error)
{
  if(!(this->rlcsa.isOk()) || !(this->rlcsa.supportsLocate()))
  {
    std::cerr << "SadaSparse::SadaSparse(): Invalid RLCSA!" << std::endl;
    return;
  }

  // In the input data, values are encoded as 1 0^value.
  std::string input_name = base_name + SADACOUNT_EXTENSION;
  std::ifstream input(input_name.c_str(), std::ios_base::binary);
  if(!input)
  {
    std::cerr << "SadaSparse::SadaSparse(): Cannot open input file " << input_name << std::endl;
    return;
  }
  usint bits = 0;
  input.read((char*)&bits, sizeof(bits));
  CSA::ReadBuffer buffer(input, bits, 1);
  input.close();

  // Build the sparse filter.
  usint total = 0, filtered_values = 0;
  sdsl::bit_vector filter_buffer(this->rlcsa.getSize() - 1, 0);
  for(usint i = 0, j = 1; i < filter_buffer.size(); i++, j++)
  {
    usint curr = 0;
    while(j < bits && !(buffer.isSet(j))) { curr++; j++; }
    if(curr > 0) { filter_buffer[i] = 1; total += curr; filtered_values++; }
  }
  this->filter = sd_vector(filter_buffer); sdsl::util::clear(filter_buffer);

  // Store the filtered values as 0^{value-1} 1.
  sdsl::sd_vector_builder builder(total, filtered_values);
  for(usint i = 0, j = 1, tail = 0; i < this->filter.size(); i++, j++)
  {
    usint curr = 0;
    while(j < bits && !(buffer.isSet(j))) { curr++; j++; }
    if(curr > 0) { tail += curr; builder.set(tail - 1); }
  }
  this->values = sd_vector(builder);

  sdsl::util::init_support(this->filter_rank, &(this->filter));
  sdsl::util::init_support(this->value_select, &(this->values));

  this->status = st_ok;
}

SadaSparse::~SadaSparse()
{
}

void
SadaSparse::writeTo(const std::string& base_name) const
{
  if(!(this->isOk()))
  {
    std::cerr << "SadaSparse::writeTo(): Status is not ok!" << std::endl;
    return;
  }

  std::string filename = base_name + SADACOUNT_EXTENSION;
  std::ofstream output(filename.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "SadaSparse::writeTo(): Cannot open output file " << filename << std::endl;
    return;
  }

  usint bits = this->filter.size() + this->values.size();
  output.write((char*)&bits, sizeof(bits));

  // Decode the uncompressed SadaCount from filter and values.
  CSA::WriteBuffer output_buffer(bits, 1);
  for(usint i = 0, j = 1, output_tail = 0, value_tail = 0; i < this->filter.size(); i++)
  {
    output_buffer.setBit(output_tail); output_tail++;
    if(this->filter[i])
    {
      usint temp = this->value_select(j) + 1; j++;
      output_tail += temp - value_tail;
      value_tail = temp;
    }
  }

  CSA::ReadBuffer* buffer = output_buffer.getReadBuffer();
  if(buffer != 0)
  {
    buffer->writeTo(output);
    delete buffer; buffer = 0;
  }

  output.close();
}

usint
SadaSparse::reportSize() const
{
  usint bytes = sizeof(this->rlcsa) + sizeof(this->status);
  bytes += sdsl::size_in_bytes(this->filter);
  bytes += sdsl::size_in_bytes(this->filter_rank);
  bytes += sdsl::size_in_bytes(this->values);
  bytes += sdsl::size_in_bytes(this->value_select);
  return bytes;
}

usint
SadaSparse::count(const std::string& pattern) const
{
  if(!(this->isOk())) { return 0; }
  pair_type sa_range = this->rlcsa.count(pattern);
  return this->count(sa_range);
}

usint
SadaSparse::count(pair_type sa_range) const
{
  if(!(this->isOk()) || sa_range.second >= this->rlcsa.getSize()) { return 0; }

  usint len = CSA::length(sa_range);
  if(len <= 1) { return len; }

  sa_range.first = this->filter_rank(sa_range.first);   // Closed lower bound for ranks of filtered values.
  sa_range.second = this->filter_rank(sa_range.second); // Open upper bound.
  if(sa_range.second > sa_range.first)
  {
    len += (sa_range.first > 0 ? this->value_select(sa_range.first) + 1 : 0);
    len -= this->value_select(sa_range.second) + 1;
  }

  return len;
}

//--------------------------------------------------------------------------

struct SadaTree;

struct SadaTreeNode
{
  union
  {
    uint string_depth;
    uint children;  // Available after binarize().
  };
  union
  {
    usint               sp; // Available until computeStoredDocs().
    std::vector<usint>* docs;
    usint               docc;
  };
  union
  {
    usint ep; // Available until computeStoredDocs().
    usint redundant;
  };

  void addChild(SadaTree* tree, SadaTreeNode* node);
  void addSibling(SadaTree* tree, SadaTreeNode* node);
  void addLeaves(SadaTree* tree);
  SadaTreeNode* addLeaf(SadaTree* tree, SadaTreeNode* left, SadaTreeNode* right, usint pos);

  // Call addLeaves() first.
  void binarize(SadaTree* tree);

  // Replaces the docs vector with document count.
  void deleteDocs();

  void insert(SadaTreeNode* from);
  void claimRedundant(SadaTreeNode* from, sc_mode_type mode);

  const static usint PARALLEL_SORT_THRESHOLD = 32768;

  // Call binarize() first.
  void computeStoredDocuments(SadaTree* tree, CSA::ReadBuffer* documents, sc_mode_type mode);

  // These are available after binarize.
  bool isReal() { return (this->children & 0x1); }
  uint childCount() { return (this->children >> 1); }
};

struct SadaTree
{
  SadaTree(usint _size) :
    data(new SadaTreeNode[_size]), nodes(0), size(_size),
    w_children(0), r_children(0), w_siblings(0), r_siblings(0)
  {
    usint item_bits = CSA::length(this->size);
    this->w_children = new CSA::WriteBuffer(this->size, item_bits);
    this->r_children = this->w_children->getReadBuffer();
    this->w_siblings = new CSA::WriteBuffer(this->size, item_bits);
    this->r_siblings = this->w_siblings->getReadBuffer();
  }

  ~SadaTree()
  {
    delete[] this->data; this->data = 0;
    delete this->w_children; this->w_children = 0;
    delete this->r_children; this->r_children = 0;
    delete this->w_siblings; this->w_siblings = 0;
    delete this->r_siblings; this->r_siblings = 0;
  }

  SadaTreeNode* getNode(uint lcp, pair_type sa_range)
  {
    SadaTreeNode* node = this->data + this->nodes;
    node->string_depth = lcp;
    node->sp = sa_range.first;
    node->ep = sa_range.second;
    this->nodes++;
    return node;
  }

  SadaTreeNode* parentOf(SadaTreeNode* left, SadaTreeNode* right)
  {
    SadaTreeNode* parent = this->getNode(2, pair_type(left->sp, right->ep));
    this->setChild(parent, left);
    this->setSibling(left, right);
    this->setSibling(right, 0);
    parent->children = parent->children << 1; // A false node.
    return parent;
  }

  void setChild(SadaTreeNode* node, SadaTreeNode* child)
  {
    this->w_children->goToItem(node - this->data);
    usint val = (child == 0 ? 0 : (child - this->data) + 1);
    this->w_children->writeItem(val);
  }

  SadaTreeNode* childOf(SadaTreeNode* node)
  {
    usint val = this->r_children->readItem(node - this->data);
    return (val > 0 ? this->data + (val - 1) : 0);
  }

  void setSibling(SadaTreeNode* node, SadaTreeNode* sibling)
  {
    this->w_siblings->goToItem(node - this->data);
    usint val = (sibling == 0 ? 0 : (sibling - this->data) + 1);
    this->w_siblings->writeItem(val);
  }

  SadaTreeNode* siblingOf(SadaTreeNode* node)
  {
    usint val = this->r_siblings->readItem(node - this->data);
    return (val > 0 ? this->data + (val - 1) : 0);
  }

  SadaTreeNode*   data;
  usint           nodes, size;

  CSA::WriteBuffer* w_children;
  CSA::ReadBuffer*  r_children;

  CSA::WriteBuffer* w_siblings;
  CSA::ReadBuffer*  r_siblings;
};

//--------------------------------------------------------------------------

void
SadaTreeNode::addChild(SadaTree* tree, SadaTreeNode* node)
{
  SadaTreeNode* child = tree->childOf(this);
  if(child == 0) { tree->setChild(this, node); return; }
  child->addSibling(tree, node);
}

void
SadaTreeNode::addSibling(SadaTree* tree, SadaTreeNode* node)
{
  SadaTreeNode* sibling = tree->siblingOf(this);
  if(sibling == 0) { tree->setSibling(this, node); return; }
  sibling->addSibling(tree, node);
}

void
SadaTreeNode::addLeaves(SadaTree* tree)
{
  std::stack<SadaTreeNode*> nodestack;
  nodestack.push(this);

  while(!(nodestack.empty()))
  {
    SadaTreeNode* curr = nodestack.top(); nodestack.pop();
    if(tree->childOf(curr) == 0 && curr->ep <= curr->sp) { continue; }
    usint expect = curr->sp;
    SadaTreeNode* prev = 0;
    for(SadaTreeNode* temp = tree->childOf(curr); temp != 0; temp = tree->siblingOf(temp))
    {
      while(expect < temp->sp) { prev = curr->addLeaf(tree, prev, temp, expect); expect++; }
      nodestack.push(temp);
      expect = temp->ep + 1;
      prev = temp;
    }
    while(expect <= curr->ep) { prev = curr->addLeaf(tree, prev, 0, expect); expect++; }
  }
}

SadaTreeNode*
SadaTreeNode::addLeaf(SadaTree* tree, SadaTreeNode* left, SadaTreeNode* right, usint pos)
{
  SadaTreeNode* leaf = tree->getNode(0, pair_type(pos, pos));
  if(left == 0) { tree->setChild(this, leaf); }
  else          { tree->setSibling(left, leaf); }
  tree->setSibling(leaf, right);
  return leaf;
}

void
SadaTreeNode::binarize(SadaTree* tree)
{
  std::vector<SadaTreeNode*> array;
  std::stack<SadaTreeNode*> nodestack;
  nodestack.push(this);

  while(!(nodestack.empty()))
  {
    SadaTreeNode* curr = nodestack.top(); nodestack.pop();
    curr->children = 0;
    array.clear();
    for(SadaTreeNode* temp = tree->childOf(curr); temp != 0; temp = tree->siblingOf(temp))
    {
      nodestack.push(temp);
      array.push_back(temp);
      curr->children++;
    }
    if(curr->children <= 2)
    {
      curr->children = (curr->children << 1) | 0x1; // A real node.
      continue;
    }

    while(curr->children > 2)
    {
      usint new_children = 0;
      for(usint i = 0; i < curr->children / 2; i++)
      {
        array[new_children] = tree->parentOf(array[2 * i], array[2 * i + 1]);
        new_children++;
      }
      if(2 * new_children < curr->children)
      {
        array[new_children] = array[curr->children - 1];
        new_children++;
      }
      curr->children = new_children;
    }
    tree->setChild(curr, array[0]); tree->setSibling(array[0], array[1]);
    curr->children = (curr->children << 1) | 0x1; // A real node.
  }
}

void
SadaTreeNode::deleteDocs()
{
  if(this->childCount() == 0) { this->docc = 1; }
  else
  {
    usint doc_count = this->docs->size();
    delete this->docs; this->docc = doc_count;
  }
}

void
SadaTreeNode::insert(SadaTreeNode* from)
{
  if(from->childCount() == 0) { this->docs->push_back(from->docc); }
  else { this->docs->insert(this->docs->end(), from->docs->begin(), from->docs->end()); }
  from->deleteDocs();
}

/*
  Claim redundant count from dummy nodes to the real node they were created from.
  When using a filter, nodes with docc == 1 do not have a redundant count.
  Mode sc_original is the original SadaCount, where the redundant counts are
  not claimed by real nodes.
*/
void
SadaTreeNode::claimRedundant(SadaTreeNode* from, sc_mode_type mode)
{
  if(from->isReal() || mode == sc_original) { return; }
  if(mode != sc_filter || from->docc > 1) { this->redundant += from->redundant; }
  from->redundant = 0;
}

void
SadaTreeNode::computeStoredDocuments(SadaTree* tree, CSA::ReadBuffer* documents, sc_mode_type mode)
{
  if(this->childCount() == 0)
  {
    this->docc = documents->readItem(this->sp);
    this->redundant = 0;
    return;
  }

  SadaTreeNode* child = tree->childOf(this);
  SadaTreeNode* sibling = tree->siblingOf(child);
  usint total_size = (child->childCount() == 0 ? 1 : child->docs->size()) +
    (sibling->childCount() == 0 ? 1 : sibling->docs->size());
  this->docs = new std::vector<usint>; this->docs->reserve(total_size);
  this->insert(child); this->insert(sibling);
  CSA::removeDuplicates(this->docs, (total_size >= PARALLEL_SORT_THRESHOLD));

  this->redundant = total_size - this->docs->size();
  this->claimRedundant(child, mode);
  this->claimRedundant(sibling, mode);
}

//--------------------------------------------------------------------------

void
buildSadaCount(const std::string& base_name, sc_mode_type mode, bool print)
{
  // Load RLCSA.
  if(print) { std::cout << "Loading RLCSA..." << std::endl; }
  RLCSA rlcsa(base_name);
  if(!(rlcsa.isOk()) || !(rlcsa.supportsLocate()))
  {
    std::cerr << "buildSadaCount(): Invalid RLCSA!" << std::endl;
    return;
  }
  rlcsa.printInfo();
  rlcsa.reportSize(true);

  // Timer start.
  double start = CSA::readTimer();

  // Build SA.
  if(print) { std::cout << "Building SA..." << std::endl; }
  usint size = rlcsa.getSize();
  usint* sa = new usint[size];
  #pragma omp parallel for schedule(static)
  for(usint i = 0; i < size; i += CSA::MEGABYTE)
  {
    rlcsa.locate(pair_type(i, std::min(i + CSA::MEGABYTE - 1, size - 1)), sa + i);
  }

  // Build LCP.
  if(print) { std::cout << "Building LCP..." << std::endl; }
  uint* lcp = new uint[size + 1]; lcp[size] = 0;
  CSA::PLCPVector* plcpvec = rlcsa.buildPLCP(16);
  CSA::PLCPVector::Iterator iter(*plcpvec);
  for(usint i = 0; i < size; i++)
  {
    lcp[i] = iter.select(sa[i]) - 2 * sa[i];
  }
  delete plcpvec; plcpvec = 0;

  // Convert SA to DA.
  if(print) { std::cout << "Building DA..." << std::endl; }
  #pragma omp parallel for schedule(static)
  for(usint i = 0; i < size; i += CSA::MEGABYTE)
  {
    rlcsa.getSequenceForPosition(sa + i, std::min(CSA::MEGABYTE, size - i));
  }
  CSA::WriteBuffer docbuffer(size, CSA::length(rlcsa.getNumberOfSequences() - 1));
  for(usint i = 0; i < size; i++) { docbuffer.writeItem(sa[i]); }
  delete[] sa; sa = 0;
  CSA::ReadBuffer* documents = docbuffer.getReadBuffer();

  // Build ST. Kind of.
  if(print) { std::cout << "Building ST..."; std::cout.flush(); }
  SadaTree* tree = new SadaTree(2 * size - 1);
  SadaTreeNode* root = tree->getNode(0, pair_type(0, 0));
  std::stack<SadaTreeNode*> nodestack; nodestack.push(root);
  SadaTreeNode* prev = 0;
  for(usint i = 1; i <= size; i++)
  {
    usint left = i - 1;
    while(lcp[i] < nodestack.top()->string_depth)
    {
      nodestack.top()->ep = i - 1;
      prev = nodestack.top(); nodestack.pop();
      left = prev->sp;
      if(lcp[i] <= nodestack.top()->string_depth) { nodestack.top()->addChild(tree, prev); prev = 0; }
    }
    if(lcp[i] > nodestack.top()->string_depth)
    {
      SadaTreeNode* curr = tree->getNode(lcp[i], pair_type(left, left));
      if(prev != 0) { curr->addChild(tree, prev); prev = 0; }
      nodestack.push(curr);
    }
  }
  while(!(nodestack.empty()))
  {
    prev = nodestack.top(); nodestack.pop();
    prev->ep = size - 1;
  }
  delete[] lcp; lcp = 0;
  root->addLeaves(tree);
  root->binarize(tree);
  if(print) { std::cout << " Size: " << tree->nodes << " / " << size << std::endl; }

  // Build the sets of document ids.
  if(print) { std::cout << "Building sets..." << std::endl; }
  for(prev = root; prev != 0; prev = tree->childOf(prev)) { nodestack.push(prev); }
  while(!(nodestack.empty()))
  {
    SadaTreeNode* curr = nodestack.top(); nodestack.pop();
    for(SadaTreeNode* temp = tree->siblingOf(curr); temp != 0; temp = tree->childOf(temp)) { nodestack.push(temp); }
    curr->computeStoredDocuments(tree, documents, mode);
  }
  root->deleteDocs();
  delete documents; documents = 0;

  // Building the counting structure.
  if(print) { std::cout << "Building bitvector..." << std::endl; }
  CSA::WriteBuffer v_buffer(2 * (size / CSA::WORD_BITS + 1));
  CSA::WriteBuffer f_buffer(2 * (size / CSA::WORD_BITS + 1));
  usint v_bits = 0, v_ones = 0;
  usint f_bits = 0, f_ones = 0;
  while(!(nodestack.empty())) { nodestack.pop(); }
  for(prev = root; prev != 0; prev = tree->childOf(prev)) { nodestack.push(prev); }
  while(!(nodestack.empty()))
  {
    SadaTreeNode* curr = nodestack.top(); nodestack.pop();
    SadaTreeNode* child = tree->childOf(curr);
    if(child != 0)
    {
      if(mode != sc_filter || curr->docc > 1)
      {
        v_buffer.setBit(v_bits);
        v_bits += 1 + curr->redundant; v_ones++;
      }
      else { f_buffer.setBit(f_bits); f_ones++; }
      f_bits++;
      for(prev = tree->siblingOf(child); prev != 0; prev = tree->childOf(prev)) { nodestack.push(prev); }
    }
  }
  delete tree; tree = 0;
  if(print)
  {
    std::cout << "  Vector: " << v_ones << " / " << v_bits << " bits" << std::endl;
    if(mode == sc_filter)
    {
      std::cout << "  Filter: " << f_ones << " / " << f_bits << " bits" << std::endl;
    }
  }

  // Timer stop.
  double total = CSA::readTimer() - start;

  // Writing results to a file.
  std::string output_name = base_name + (mode == sc_filter ? SADAFILTER_EXTENSION : SADACOUNT_EXTENSION);
  std::ofstream output(output_name.c_str(), std::ios_base::binary);
  if(!output)
  {
    std::cerr << "buildSadaCount(): Cannot open output file " << output_name << std::endl;
    return;
  }
  {
    output.write((char*)&v_bits, sizeof(v_bits));
    CSA::ReadBuffer* vec = v_buffer.getReadBuffer(v_bits, 1);
    vec->writeTo(output);
    if(mode == sc_export) { vec->exportTo(base_name + SADAEXPORT_EXTENSION); }
    delete vec; vec = 0;
  }
  if(mode == sc_filter)
  {
    output.write((char*)&f_bits, sizeof(f_bits));
    CSA::ReadBuffer* filter = f_buffer.getReadBuffer(f_bits, 1);
    filter->writeTo(output);
    delete filter; filter = 0;
  }
  output.close();

  if(print)
  {
    double megabytes = size / (double)CSA::MEGABYTE;
    std::cout << "Processed " << megabytes << " megabytes in " << total << " seconds (" << (megabytes / total) << " MB/s)." << std::endl;
    std::cout << "Memory: " << CSA::memoryUsage() << " kB" << std::endl;
    std::cout << std::endl;
  }
}

//--------------------------------------------------------------------------
